@@ -20,18 +20,20 @@ from ._frontend import module_name, module_version
 
 from .utils import binary_image, populate_args, image_bytes_to_array, commands_to_buffer
 
-COMMANDS = {
-    'fillRect': 0, 'strokeRect': 1, 'fillRects': 2, 'strokeRects': 3, 'clearRect': 4, 'fillArc': 5,
-    'fillCircle': 6, 'strokeArc': 7, 'strokeCircle': 8, 'fillArcs': 9, 'strokeArcs': 10,
-    'fillCircles': 11, 'strokeCircles': 12, 'strokeLine': 13, 'beginPath': 14, 'closePath': 15,
-    'stroke': 16, 'fillPath': 17, 'fill': 18, 'moveTo': 19, 'lineTo': 20,
-    'rect': 21, 'arc': 22, 'ellipse': 23, 'arcTo': 24, 'quadraticCurveTo': 25,
-    'bezierCurveTo': 26, 'fillText': 27, 'strokeText': 28, 'setLineDash': 29, 'drawImage': 30,
-    'putImageData': 31, 'clip': 32, 'save': 33, 'restore': 34, 'translate': 35,
-    'rotate': 36, 'scale': 37, 'transform': 38, 'setTransform': 39, 'resetTransform': 40,
-    'set': 41, 'clear': 42, 'sleep': 43, 'fillPolygon': 44, 'strokePolygon': 45,
-    'strokeLines': 46,
-}
+_CMD_LIST = ['fillRect', 'strokeRect', 'fillRects', 'strokeRects', 'clearRect', 'fillArc',
+             'fillCircle', 'strokeArc', 'strokeCircle', 'fillArcs', 'strokeArcs',
+             'fillCircles', 'strokeCircles', 'strokeLine', 'beginPath', 'closePath',
+             'stroke', 'fillPath', 'fill', 'moveTo', 'lineTo',
+             'rect', 'arc', 'ellipse', 'arcTo', 'quadraticCurveTo',
+             'bezierCurveTo', 'fillText', 'strokeText', 'setLineDash', 'drawImage',
+             'putImageData', 'clip', 'save', 'restore', 'translate',
+             'rotate', 'scale', 'transform', 'setTransform', 'resetTransform',
+             'set', 'clear', 'sleep', 'fillPolygon', 'strokePolygon',
+             'strokeLines', 'fillPolygons', 'strokePolygons', 'strokeLineSegments',
+             'fillStyledRects', 'strokeStyledRects', 'fillStyledCircles', 'strokeStyledCircles',
+             'fillStyledArcs', 'strokeStyledArcs', 'fillStyledPolygons',
+             'strokeStyledPolygons', 'strokeStyledLineSegments']
+COMMANDS = {v: i for i, v in enumerate(_CMD_LIST)}
 
 
 # Traitlets does not allow validating without creating a trait class, so we need this
@@ -54,6 +56,55 @@ def _validate_number(value, min_val, max_val):
     raise TraitError('{} is not in the range [{}, {}]'.format(value, min_val, max_val))
 
 
+def _serialize_list_of_polygons_or_linestrokes(points, points_per_item, item_name, min_elements):
+    if isinstance(points, list):
+        if points_per_item is not None:
+            raise RuntimeError("when points are a list, points_per_item must be None")
+        points_per_item = []
+        np_polygons = []
+        for i, polygon_points in enumerate(points):
+            polygon_points = np.require(polygon_points, requirements=['C'])
+            if polygon_points.shape[1] != 2:
+                raise RuntimeError(f"item {i} in points have wrong shape: `{polygon_points.shape}` but must be of type (n,2)")
+            points_per_item.append(polygon_points.shape[0])
+            np_polygons.append(polygon_points.ravel())
+
+        num_polygons = len(points)
+        flat_points = np.concatenate(np_polygons)
+        points_per_item = np.array(points_per_item)
+
+    elif isinstance(points, np.ndarray):
+        points = np.require(points, requirements=['C'])
+        shape = points.shape
+        ndim = points.ndim
+
+        if ndim <= 2:
+            if points_per_item is None:
+                raise RuntimeError("when points are given as a 1d / 2d array, points_per_item must not be None")
+            if ndim == 1:
+                flat_points = points
+            elif ndim == 2:
+                if shape[1] != 2:
+                    raise RuntimeError(f"points have wrong shape: `{shape}`. When points are given as a 2D array the shape must be of (n,2)")
+                flat_points = points.ravel()
+            num_polygons = len(points_per_item)
+        elif ndim == 3:
+            if points_per_item is not None:
+                raise RuntimeError("when points are a list, points_per_item must be None")
+            if shape[2] != 2:
+                raise RuntimeError(f"Points have wrong shape: `{shape}`: When points are given as a 3D array the shape must be of (n_{item_name}, n_points_per_{item_name}, 2)")
+            if shape[1] < min_elements:
+                raise RuntimeError(f"Points have wrong shape: `{shape}`: when points are given as a 3D array the shape must be of (n_{item_name}, n_points_per_{item_name}, 2) and n_points_per_{item_name} must be >= {min_elements} ")
+            flat_points = points.ravel()
+            num_polygons = shape[0]
+            points_per_item = shape[1]
+        else:
+            raise RuntimeError("ndarray must have ndim <= 3")
+    else:
+        raise RuntimeError("points must be a list or an ndarray")
+    return num_polygons, flat_points, points_per_item
+
+
 class Path2D(Widget):
     """Create a Path2D.
 
@@ -254,6 +305,8 @@ class Canvas(_CanvasBase):
     _model_name = Unicode('CanvasModel').tag(sync=True)
     _view_name = Unicode('CanvasView').tag(sync=True)
 
+    _send_client_ready_event = Bool(True).tag(sync=True)
+
     #: (valid HTML color or Gradient or Pattern) The color for filling rectangles and paths. Default to ``'black'``.
     fill_style = Union((Color(), Instance(_CanvasGradient), Instance(Pattern)), default_value='black')
 
@@ -466,6 +519,54 @@ class Canvas(_CanvasBase):
 
         self._send_canvas_command(COMMANDS['strokeRects'], args, buffers)
 
+    def fill_styled_rects(self, x, y, width, height, color, alpha=1):
+        """Draw filled and styled rectangles of sizes ``(width, height)`` at the ``(x, y)`` positions
+
+        Where ``x``, ``y``, ``width`` and ``height`` arguments are NumPy arrays, lists or scalar values.
+        If ``height`` is None, it is set to the same value as width.
+        ``color`` is an (n_rect x 3) NumPy array with the colors and ``alpha`` is an (n_rect) NumPy array
+        with the alpha channel values.
+        """
+        args = []
+        buffers = []
+
+        populate_args(x, args, buffers)
+        populate_args(y, args, buffers)
+        populate_args(width, args, buffers)
+
+        if height is None:
+            args.append(args[-1])
+        else:
+            populate_args(height, args, buffers)
+
+        populate_args(color, args, buffers)
+        populate_args(alpha, args, buffers)
+        self._send_canvas_command(COMMANDS['fillStyledRects'], args, buffers)
+
+    def stroke_styled_rects(self, x, y, width, height, color, alpha=1):
+        """Draw rectangular styled outlines of sizes ``(width, height)`` at the ``(x, y)`` positions.of sizes ``(width, height)``
+
+        Where ``x``, ``y``, ``width`` and ``height`` arguments are NumPy arrays, lists or scalar values.
+        If ``height`` is None, it is set to the same value as width.
+        ``color`` is an (n_rect x 3) NumPy array with the colors and ``alpha`` is an (n_rect) NumPy array
+        with the alpha channel values.
+        """
+        args = []
+        buffers = []
+
+        populate_args(x, args, buffers)
+        populate_args(y, args, buffers)
+        populate_args(width, args, buffers)
+
+        if height is None:
+            args.append(args[-1])
+        else:
+            populate_args(height, args, buffers)
+
+        populate_args(color, args, buffers)
+        populate_args(alpha, args, buffers)
+        self._send_canvas_command(COMMANDS['strokeStyledRects'], args, buffers)
+
     def clear_rect(self, x, y, width, height=None):
         """Clear the specified rectangular area of size ``(width, height)`` at the ``(x, y)`` position, making it fully transparent."""
         if height is None:
@@ -552,6 +653,76 @@ class Canvas(_CanvasBase):
 
         self._send_canvas_command(COMMANDS['strokeCircles'], args, buffers)
 
+    def fill_styled_circles(self, x, y, radius, color, alpha=1):
+        """Draw a filled circles centered at ``(x, y)`` with a radius of ``radius``.
+
+        Where ``x``, ``y``, ``radius``  and ``alpha`  are NumPy arrays, lists or scalar values.
+        ``color`` must be an (nx3) NumPy arrays with the color for each circle.
+        """
+        args = []
+        buffers = []
+
+        populate_args(x, args, buffers)
+        populate_args(y, args, buffers)
+        populate_args(radius, args, buffers)
+        populate_args(color, args, buffers)
+        populate_args(alpha, args, buffers)
+        self._send_canvas_command(COMMANDS['fillStyledCircles'], args, buffers)
+
+    def stroke_styled_circles(self, x, y, radius, color, alpha=1):
+        """Draw filled circles centered at ``(x, y)`` with a radius of ``radius``.
+
+        Where ``x``, ``y``, ``radius``  and ``alpha``  are NumPy arrays, lists or scalar values.
+        ``color`` must be an nx3 NumPy arrays with the color for each circle
+        """
+        args = []
+        buffers = []
+
+        populate_args(x, args, buffers)
+        populate_args(y, args, buffers)
+        populate_args(radius, args, buffers)
+        populate_args(color, args, buffers)
+        populate_args(alpha, args, buffers)
+        self._send_canvas_command(COMMANDS['strokeStyledCircles'], args, buffers)
+
+    def fill_styled_arcs(self, x, y, radius, start_angle, end_angle, color, alpha=1, anticlockwise=False):
+        """Draw filled and styled arcs centered at ``(x, y)`` with a radius of ``radius``.
+
+        Where ``x``, ``y``, ``radius`` and other arguments are NumPy arrays, lists or scalar values.
+        """
+        args = []
+        buffers = []
+
+        populate_args(x, args, buffers)
+        populate_args(y, args, buffers)
+        populate_args(radius, args, buffers)
+        populate_args(start_angle, args, buffers)
+        populate_args(end_angle, args, buffers)
+        args.append(anticlockwise)
+        populate_args(color, args, buffers)
+        populate_args(alpha, args, buffers)
+
+        self._send_canvas_command(COMMANDS['fillStyledArcs'], args, buffers)
+
+    def stroke_styled_arcs(self, x, y, radius, start_angle, end_angle, color, alpha=1, anticlockwise=False):
+        """Draw an styled arc outlines centered at ``(x, y)`` with a radius of ``radius``.
+
+        Where ``x``, ``y``, ``radius`` and other arguments are NumPy arrays, lists or scalar values.
+        """
+        args = []
+        buffers = []
+
+        populate_args(x, args, buffers)
+        populate_args(y, args, buffers)
+        populate_args(radius, args, buffers)
+        populate_args(start_angle, args, buffers)
+        populate_args(end_angle, args, buffers)
+        args.append(anticlockwise)
+        populate_args(color, args, buffers)
+        populate_args(alpha, args, buffers)
+
+        self._send_canvas_command(COMMANDS['strokeStyledArcs'], args, buffers)
+
     # Polygon methods
     def fill_polygon(self, points):
         """Fill a polygon from a list of points ``[(x1, y1), (x2, y2), ..., (xn, yn)]``."""
@@ -571,6 +742,115 @@ class Canvas(_CanvasBase):
 
         self._send_canvas_command(COMMANDS['strokePolygon'], args, buffers)
 
+    def fill_polygons(self, points, points_per_polygon=None):
+        """" Draw many filled polygons at once:
+
+            Args:
+                points (list or ndarray): The polygons points:
+
+                    The points can be specified as list or as ndarray:
+                    If the points are a list it must a an be a list of ndarrays,
+                    where each ndarray is a nx2 array of coordinates
+                    (n can be different for each entry)
+                    If the points are given as ndarray it must be either:
+
+                    - a 3d array: the shape of the array is (n_polyons, n_points_per_polygon, 2)
+                    - a 2d array: the shape of the array is (n,  2) and in additional  `points_per_polygon`
+                      must be specified st. we know the the number of points for each individual polygon.
+                      Note that the number of points in ``points`` must match the points_per_polygon.
+                      array: ie: `np.sum(points_per_polygon) == points.shape[0]`
+
+                points_per_polygon (ndarray):
+                    ndarray with number of points for each polygon. Must **only** be given if points are
+                    given as `flat` 2D array.
+
+        """
+        self._draw_polygons_or_linesegments('fillPolygons', points, None, None, points_per_polygon, False, 3, "polygon")
+
+    def stroke_polygons(self, points, points_per_polygon=None):
+        """" Draw many stroked polygons at once:
+
+            Args:
+                points (list or ndarray): The polygons points:
+
+                    The points can be specified as list or as ndarray:
+                    If the points are a list it must a an be a list of ndarrays,
+                    where each ndarray is a nx2 array of coordinates
+                    (n can be different for each entry)
+                    If the points are given as ndarray it must be either:
+
+                    - a 3d array: the shape of the array is (n_polyons, n_points_per_polygon, 2)
+                    - a 2d array: the shape of the array is (n,  2) and in additional  `points_per_polygon`
+                      must be specified st. we know the the number of points for each individual polygon.
+                      Note that the number of points in ``points`` must match the points_per_polygon.
+                      array: ie: `np.sum(points_per_polygon) == points.shape[0]`
+
+                points_per_polygon (ndarray):
+                    ndarray with number of points for each polygon. Must **only** be given if points are
+                    given as `flat` 2D array.
+        """
+        self._draw_polygons_or_linesegments('strokePolygons', points, None, None, points_per_polygon, False, 3, "polygon")
+
+    def fill_styled_polygons(self, points, color, alpha=1, points_per_polygon=None):
+        """" Draw many filled polygons at once:
+
+            Args:
+                points (list or ndarray): The polygons points:
+
+                    The points can be specified as list or as ndarray:
+                    If the points are a list it must a an be a list of ndarrays,
+                    where each ndarray is a nx2 array of coordinates
+                    (n can be different for each entry)
+                    If the points are given as ndarray it must be either:
+
+                    - a 3d array: the shape of the array is (n_polyons, n_points_per_polygon, 2)
+                    - a 2d array: the shape of the array is (n,  2) and in additional  `points_per_polygon`
+                      must be specified st. we know the the number of points for each individual polygon.
+                      Note that the number of points in ``points`` must match the points_per_polygon.
+                      array: ie: `np.sum(points_per_polygon) == points.shape[0]`
+
+                color (ndarray)
+                    An (n_polyons,3) array with the color for each polygon
+                alpha (ndarray,list,scalar):
+                    An  array with the alpha value for each polygon. Can be a scalar and the
+                    same value is used for all polygons
+                points_per_polygon (ndarray):
+                    ndarray with number of points for each polygon. Must **only** be given if points are
+                    given as `flat` 2D array.
+
+        """
+        self._draw_polygons_or_linesegments('fillStyledPolygons', points, color, alpha, points_per_polygon, True, 3, "polygon")
+
+    def stroke_styled_polygons(self, points, color, alpha=1, points_per_polygon=None):
+        """" Draw many stroked polygons at once:
+
+            Args:
+                points (list or ndarray): The polygons points:
+
+                    The points can be specified as list or as ndarray:
+                    If the points are a list it must a an be a list of ndarrays,
+                    where each ndarray is a nx2 array of coordinates
+                    (n can be different for each entry)
+                    If the points are given as ndarray it must be either:
+
+                    - a 3d array: the shape of the array is (n_polyons, n_points_per_polygon, 2)
+                    - a 2d array: the shape of the array is (n,  2) and in additional  `points_per_polygon`
+                      must be specified st. we know the the number of points for each individual polygon.
+                      Note that the number of points in ``points`` must match the points_per_polygon.
+                      array: ie: `np.sum(points_per_polygon) == points.shape[0]`
+
+                color (ndarray)
+                    An (n_polyons,3) array with the color for each polygon
+                alpha (ndarray,list,scalar):
+                    An  array with the alpha value for each polygon. Can be a scalar and the
+                    same value is used for all polygons
+                points_per_polygon (ndarray):
+                    ndarray with number of points for each polygon. Must **only** be given if points are
+                    given as `flat` 2D array.
+
+        """
+        self._draw_polygons_or_linesegments('strokeStyledPolygons', points, color, alpha, points_per_polygon, True, 3, "polygon")
+
     # Lines methods
     def stroke_line(self, x1, y1, x2, y2):
         """Draw a line from ``(x1, y1)`` to ``(x2, y2)``."""
@@ -585,6 +865,60 @@ class Canvas(_CanvasBase):
 
         self._send_canvas_command(COMMANDS['strokeLines'], args, buffers)
 
+    def stroke_styled_line_segments(self, points, color, alpha=1, points_per_line_segment=None):
+        """ Draw many line segments at once:
+
+            Args:
+                points (list or ndarray): The line_segments points:
+
+                    The points can be specified as list or as ndarray:
+                    If the points are a list it must a an be a list of ndarrays,
+                    where each ndarray is a nx2 array of coordinates
+                    (n can be different for each entry)
+                    If the points are given as ndarray it must be either:
+
+                    - a 3d array: the shape of the array is (n_line_segments, n_points_per_polygon, 2)
+                    - a 2d array: the shape of the array is (n,  2) and in additional  `points_per_line_segment`
+                        must be specified st. we know the the number of points for each individual line_segment.
+                        Note that the number of points in ``points`` must match the points_per_line_segment.
+                        array: ie: `np.sum(points_per_line_segment) == points.shape[0]`
+
+                color (ndarray)
+                    An (n_line_segments,3) array with the color for each line_segment
+                alpha (ndarray,list,scalar):
+                    An  array with the alpha value for each line_segment. Can be a scalar and the
+                    same value is used for all line_segments
+                points_per_line_segment (ndarray):
+                    ndarray with number of points for each line_segment. Must **only** be given if points are
+                    given as `flat` 2D array.
+
+        """
+        self._draw_polygons_or_linesegments('strokeStyledLineSegments', points, color, alpha, points_per_line_segment, True, 2, "line_segment")
+
+    def stroke_line_segments(self, points, points_per_line_segment=None):
+        """ Draw many stroked line_segments at once:
+
+            Args:
+                points (list or ndarray): The line_segments points:
+
+                    The points can be specified as list or as ndarray:
+                    If the points are a list it must a an be a list of ndarrays,
+                    where each ndarray is a nx2 array of coordinates
+                    (n can be different for each entry)
+                    If the points are given as ndarray it must be either:
+
+                    - a 3d array: the shape of the array is (n_line_segments, n_points_per_polygon, 2)
+                    - a 2d array: the shape of the array is (n,  2) and in additional  `points_per_line_segment `
+                      must be specified st. we know the the number of points for each individual polygon.
+                      Note that the number of points in ``points`` must match the points_per_line_segment .
+                      array: ie: `np.sum(points_per_line_segment  ) == points.shape[0]`\
+
+                points_per_line_segment  (ndarray):
+                    ndarray with number of points for each polygon. Must **only** be given if points are
+                    given as `flat` 2D array.
+        """
+        self._draw_polygons_or_linesegments('strokeLineSegments', points, None, None, points_per_line_segment, False, 2, "line_segment")
+
     # Paths methods
     def begin_path(self):
         """Call this method when you want to create a new path."""
@@ -887,6 +1221,26 @@ class Canvas(_CanvasBase):
         if content.get('event', '') == 'touch_cancel':
             self._touch_cancel_callbacks([(touch['x'], touch['y']) for touch in content['touches']])
 
+    def _draw_polygons_or_linesegments(self, cmd, points, color, alpha, points_per_item, with_style, min_elements, item_name):
+        args = []
+        buffers = []
+
+        num_polygons, flat_points, points_per_item = _serialize_list_of_polygons_or_linestrokes(
+            points=points, points_per_item=points_per_item, item_name=item_name, min_elements=min_elements)
+
+        if with_style:
+            color = np.require(color, requirements=['C'], dtype='uint8')
+            if color.ndim != 1:
+                color = color .ravel()
+
+        populate_args(num_polygons, args, buffers)
+        populate_args(flat_points, args, buffers)
+        populate_args(points_per_item, args, buffers)
+        if with_style:
+            populate_args(color, args, buffers)
+            populate_args(alpha, args, buffers)
+        self._send_canvas_command(COMMANDS[cmd], args, buffers)
+
 
 class RoughCanvas(Canvas):
     """Create a RoughCanvas widget. It gives a hand-drawn-like style to your drawings.
