@@ -72,7 +72,10 @@ const COMMANDS = [
   'putImageData', 'clip', 'save', 'restore', 'translate',
   'rotate', 'scale', 'transform', 'setTransform', 'resetTransform',
   'set', 'clear', 'sleep', 'fillPolygon', 'strokePolygon',
-  'strokeLines',
+  'strokeLines','fillPolygons','strokePolygons','strokeLineSegments',
+  'fillStyledRects', 'strokeStyledRects', 'fillStyledCircles','strokeStyledCircles',
+  'fillStyledArcs', 'strokeStyledArcs', 'fillStyledPolygons',
+  'strokeStyledPolygons','strokeStyledLineSegments',
 ];
 
 
@@ -244,6 +247,7 @@ class CanvasModel extends DOMWidgetModel {
       height: 500,
       sync_image_data: false,
       image_data: null,
+      _send_client_ready_event: true,
     };
   }
 
@@ -275,7 +279,9 @@ class CanvasModel extends DOMWidgetModel {
     this.on('change:sync_image_data', this.syncImageData.bind(this));
     this.on('msg:custom', this.onCommand.bind(this));
 
-    this.send({ event: 'client_ready' }, {});
+    if (this.get('_send_client_ready_event')) {
+      this.send({ event: 'client_ready' }, {});
+    }
   }
 
   private async drawImageData() {
@@ -388,6 +394,42 @@ class CanvasModel extends DOMWidgetModel {
       case 'clear':
         this.clearCanvas();
         break;
+      case 'fillPolygons':
+        await this.drawPolygonOrLineSegments(args, buffers, true, true)
+        break
+      case 'strokePolygons':
+        await this.drawPolygonOrLineSegments(args, buffers, false, true)
+        break
+      case 'strokeLineSegments':
+        await this.drawPolygonOrLineSegments(args, buffers, false, false)
+        break
+      case 'fillStyledRects':
+        await this.drawStyledRects(args, buffers, true)
+        break;
+      case 'strokeStyledRects':
+        await this.drawStyledRects(args, buffers, false)
+        break;
+      case 'fillStyledCircles':
+        await this.drawStyledCircles(args, buffers, true)
+        break;
+      case 'strokeStyledCircles':
+        await this.drawStyledCircles(args, buffers, false)
+        break;
+      case 'fillStyledArcs':
+        await this.drawStyledArcs(args, buffers, true)
+        break;
+      case 'strokeStyledArcs':
+        await this.drawStyledArcs(args, buffers, false)
+        break;
+      case 'fillStyledPolygons':
+        await this.drawStyledPolygonOrLineSegments(args, buffers, true, true)
+        break
+      case 'strokeStyledPolygons':
+        await this.drawStyledPolygonOrLineSegments(args, buffers, false, true)
+        break
+      case 'strokeStyledLineSegments':
+        await this.drawStyledPolygonOrLineSegments(args, buffers, false, false)
+        break
       default:
         this.executeCommand(name, args);
         break;
@@ -425,6 +467,30 @@ class CanvasModel extends DOMWidgetModel {
       callback(x.getItem(idx), y.getItem(idx), width.getItem(idx), height.getItem(idx));
     }
   }
+  private drawStyledRects(args: any[], buffers: any, fill: boolean){
+    const x = getArg(args[0], buffers);
+    const y = getArg(args[1], buffers);
+    const width = getArg(args[2], buffers);
+    const height = getArg(args[3], buffers);
+    const colors = getArg(args[4], buffers);
+    const alpha = getArg(args[5], buffers);
+
+    const numberRects = Math.min(x.length, y.length,  width.length, height.length);
+    this.ctx.save()
+    for (let idx = 0; idx < numberRects; ++idx) {
+        // get color for this circle
+        const ci = 3*idx
+        const color = `rgba(${colors.getItem(ci)}, ${colors.getItem(ci+1)}, ${colors.getItem(ci+2)}, ${alpha.getItem(idx)})`;
+        this.setStyle(color, fill)
+        if(fill)
+        {
+          this.fillRect(x.getItem(idx), y.getItem(idx), width.getItem(idx), height.getItem(idx));
+        }else{
+          this.strokeRect(x.getItem(idx), y.getItem(idx), width.getItem(idx), height.getItem(idx));
+        }
+    }
+    this.ctx.restore()
+  }
 
   protected fillArc(x: number, y: number, radius: number, startAngle: number, endAngle: number, anticlockwise: boolean) {
     this.ctx.beginPath();
@@ -469,6 +535,7 @@ class CanvasModel extends DOMWidgetModel {
     }
   }
 
+
   protected fillCircle(x: number, y: number, radius: number) {
     this.ctx.beginPath();
     this.ctx.arc(x, y, radius, 0, 2 * Math.PI);
@@ -497,6 +564,155 @@ class CanvasModel extends DOMWidgetModel {
     }
   }
 
+
+  private setStyle(style:any, fill:boolean){
+    if(fill){
+      this.ctx.fillStyle  = style
+    }else{
+      this.ctx.strokeStyle  = style
+    }
+  }
+
+  private drawStyledCircles(args: any[], buffers: any, fill: boolean){
+    const x = getArg(args[0], buffers);
+    const y = getArg(args[1], buffers);
+    const radius = getArg(args[2], buffers);
+    const colors = getArg(args[3], buffers);
+    const alpha = getArg(args[4], buffers);
+
+    const numberCircles = Math.min(x.length, y.length, radius.length)
+    this.ctx.save()
+    for (let idx = 0; idx < numberCircles; ++idx) {
+        // get color for this circle
+        const ci = 3*idx
+        const color = `rgba(${colors.getItem(ci)}, ${colors.getItem(ci+1)}, ${colors.getItem(ci+2)}, ${alpha.getItem(idx)})`;
+        this.setStyle(color, fill)
+        if(fill)
+        {
+          this.fillCircle(x.getItem(idx), y.getItem(idx), radius.getItem(idx))
+        }else{
+          this.strokeCircle(x.getItem(idx), y.getItem(idx), radius.getItem(idx))
+        }
+    }
+    this.ctx.restore()
+  }
+
+ private drawStyledArcs(args: any[], buffers: any, fill: boolean){
+    const x = getArg(args[0], buffers);
+    const y = getArg(args[1], buffers);
+    const radius = getArg(args[2], buffers);
+    const startAngle = getArg(args[3], buffers);
+    const endAngle = getArg(args[4], buffers);
+    const anticlockwise = getArg(args[5], buffers);
+    const colors = getArg(args[6], buffers);
+    const alpha = getArg(args[7], buffers);
+
+    const numberArcs = Math.min(
+      x.length, y.length, radius.length,
+      startAngle.length, endAngle.length
+    );
+    this.ctx.save()
+    for (let idx = 0; idx < numberArcs; ++idx) {
+        // get color for this circle
+        const ci = 3*idx
+        const color = `rgba(${colors.getItem(ci)}, ${colors.getItem(ci+1)}, ${colors.getItem(ci+2)}, ${alpha.getItem(idx)})`;
+        this.setStyle(color, fill)
+        if(fill)
+        {
+          this.fillArc(x.getItem(idx), y.getItem(idx), radius.getItem(idx),
+            startAngle.getItem(idx), endAngle.getItem(idx),
+            anticlockwise.getItem(idx));
+        }else{
+          this.strokeArc(x.getItem(idx), y.getItem(idx), radius.getItem(idx),
+            startAngle.getItem(idx), endAngle.getItem(idx),
+            anticlockwise.getItem(idx));
+        }
+    }
+    this.ctx.restore()
+  }
+
+
+  private drawStyledPolygonOrLineSegments(args: any[], buffers: any, fill: boolean, close: boolean){
+
+    // a scalar
+    const numPolygons = args[0];
+
+    // always array 
+    const points = getArg(args[1], buffers);
+
+    // array or scalar
+    const sizes = getArg(args[2], buffers);
+
+    // always array
+    const colors = getArg(args[3], buffers);
+
+    // array or scalar
+    const alpha = getArg(args[4], buffers);
+
+    this.ctx.save()
+
+    var start : number  = 0
+    for (let idx = 0; idx < numPolygons; ++idx) {
+        // get color for this circle
+        const ci = 3*idx
+        const color = `rgba(${colors.getItem(ci)}, ${colors.getItem(ci+1)}, ${colors.getItem(ci+2)}, ${alpha.getItem(idx)})`;
+        this.setStyle(color, fill)
+
+        // start / stop in the points array fr this polygon
+        const size = sizes.getItem(idx) * 2;
+        const stop = start + size;
+
+        // Move to the first point, then create lines between points
+        this.ctx.beginPath();
+        this.ctx.moveTo(points.getItem(start), points.getItem(start+1));
+
+        // draw all points of the polygon (except start) 
+        for(let idp = start+2; idp < stop; idp += 2){
+          this.ctx.lineTo(points.getItem(idp), points.getItem(idp + 1));
+        }
+        start = stop
+        if(close){
+          this.ctx.closePath();
+        }
+        fill ? this.ctx.fill() : this.ctx.stroke()
+    }
+    this.ctx.restore()
+  }
+
+  private drawPolygonOrLineSegments(args: any[], buffers: any, fill: boolean, close: boolean){
+
+    // a scalar
+    const numPolygons = args[0];
+
+    // always array 
+    const points = getArg(args[1], buffers);
+
+    // array or scalar
+    const sizes = getArg(args[2], buffers);
+
+    var start : number  = 0
+    for (let idx = 0; idx < numPolygons; ++idx) {
+
+        // start / stop in the points array fr this polygon
+        const size = sizes.getItem(idx) * 2;
+        const stop = start + size;
+
+        // Move to the first point, then create lines between points
+        this.ctx.beginPath();
+        this.ctx.moveTo(points.getItem(start), points.getItem(start+1));
+
+        // draw all points of the polygon (except start) 
+        for(let idp = start+2; idp < stop; idp += 2){
+          this.ctx.lineTo(points.getItem(idp), points.getItem(idp + 1));
+        }
+        start = stop
+        if(close){
+          this.ctx.closePath();
+        }
+        fill ? this.ctx.fill() : this.ctx.stroke()
+    }
+  }
+
   protected strokeLine(args: any[], buffers: any) {
     this.ctx.beginPath();
     this.ctx.moveTo(args[0], args[1]);
